# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Utilities."""

from __future__ import absolute_import, print_function

from datetime import timedelta
from functools import partial

from flask import current_app, render_template, request
from flask_babelex import gettext as _
from flask_mail import Message
from invenio_mail.tasks import send_email
from invenio_pidstore.resolver import Resolver
from invenio_records.api import Record
from invenio_records_rest.utils import LazyPIDValue, obj_or_import_string
from six.moves.urllib.parse import parse_qs, urlencode, urlsplit, urlunsplit
from werkzeug.routing import BaseConverter
from werkzeug.utils import cached_property


def format_url_template(url_template, absolute=True, **kwargs):
    """Formats a URL template-like string."""
    scheme, netloc, path, query, fragment = urlsplit(url_template)
    scheme = scheme or request.scheme
    netloc = netloc or request.host

    # TODO: Find a more specific exception to raise here (or custom)
    if absolute:
        assert netloc

    path = path.format(**kwargs)
    query = urlencode(
        {k: v.format(**kwargs) for k, v in parse_qs(query).items()})
    return urlunsplit((scheme, netloc, path, query, fragment))


def send_invitation_email(membership_request, recipients, community):
    """Send email notification for a member invitation."""
    msg = Message(
        _("Community membership invitation"),
        sender=current_app.config.get('SUPPORT_EMAIL'),
        recipients=recipients,
    )
    template = ("invitation_email.tpl" if membership_request.is_invite
                else "member_request_email.tpl")
    msg.body = render_template(
        "invenio_communities/{}".format(template),
        community=community,
        days=timedelta(
            seconds=current_app.config[
                "COMMUNITIES_MEMBERSHIP_REQUESTS_CONFIRMLINK_EXPIRES_IN"]
        ).days,
        membership_request=membership_request,
        invitation_link=format_url_template(
            # TODO: change to use value from config
            '/communities/members/requests/{req_id}',
            req_id=str(membership_request.id),
        )
    )
    send_email.delay(msg.__dict__)


# TODO: Remove when
# https://github.com/inveniosoftware/invenio-records-rest/pull/294 is merged
class LazyPIDConverter(BaseConverter):
    """Converter for PID values in the route mapping."""

    def __init__(self, url_map, pid_type, getter=None, record_class=None,
                 object_type='rec'):
        """Initialize the converter."""
        super(LazyPIDConverter, self).__init__(url_map)
        self.pid_type = pid_type
        self.getter = getter
        self.record_class = record_class
        self.object_type = object_type

    @cached_property
    def resolver(self):
        """PID resolver."""
        record_cls = obj_or_import_string(
            self.record_class, default=Record)
        getter = obj_or_import_string(
            self.getter,
            default=partial(record_cls.get_record, with_deleted=True))
        return Resolver(
            pid_type=self.pid_type,
            object_type=self.object_type,
            getter=getter
        )

    def to_python(self, value):
        """Resolve PID value."""
        return LazyPIDValue(self.resolver, value)


comid_url_converter = (
    'lazy_pid(comid,'
    'record_class="invenio_communities.api:Community",'
    'object_type="com")'
)




# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community records indexing receivers."""

from invenio_communities.records.api import Record, RecordCommunitiesCollection


def record_indexer_receiver(sender, json=None, record=None, **kwargs):
    """Add communities information to record.

    To integrate this indexer you need to add to your record's ES mapping the
    following ``_communities`` property:

    ..code-block:: json

    {
        "_communities": {
            "type": "object",
            "properties": {
                "accepted": {
                    "type": "object",
                    "properties": {
                        "id": {
                            "type": "keyword",
                        },
                        "comid": {
                            "type": "keyword",
                        }
                        "title": {
                            "type": "text",
                        },
                        "request_id": {
                            "type": "keyword",
                        }
                        "created_by": {
                            "type": "integer",
                        }
                    }
                },
                "pending": {
                    <same structure as accepted>
                },
                "rejected": {
                    <same structure as accepted>
                }
            }
        }
    }
    """
    # TODO: Remove when the PID mixin is in the base record class
    _record = Record(record, model=record.model)
    json['_communities'] = RecordCommunitiesCollection(_record).as_dict()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Records API."""

from __future__ import absolute_import, print_function

from invenio_communities.members.models import CommunityMember, \
    CommunityMetadata, MembershipRequest
from invenio_communities.utils import send_invitation_email


class CommunityMembersAPI(object):

    @classmethod
    def invite_member(cls, community, email, role, send_email=True):
        membership_request = MembershipRequest.create(
            community.id, True, role=role, email=email)
        if send_email:
            send_invitation_email(membership_request, [email], community)
        return membership_request

    @classmethod
    def join_community(cls, user_id, email, community, send_email=True):
        membership_request = MembershipRequest.create(
            community.id, False, user_id=user_id, email=email)
        if send_email:
            community_admins = CommunityMember.get_admins(community.id)
            admin_emails = [admin.email for admin in community_admins]
            send_invitation_email(
                membership_request, admin_emails, community)
        return membership_request

    @classmethod
    def get_members(cls, comm_id):
        return CommunityMember.get_members(comm_id)

    @classmethod
    def delete_member(cls, comm_id, user_id):
        CommunityMember.delete(comm_id, user_id)

    @classmethod
    def set_default_admin(cls, community):
        CommunityMember.create(
            comm_id=community.id,
            user_id=community.json['created_by'],
            role='A')

    @classmethod
    def modify_membership(cls, user_id, comm_id, role):
        member = CommunityMember.query.filter_by(
            comm_id=comm_id, user_id=user_id).one_or_none()
        if member:
            member.role = role
            return member
        else:
            return None

    # TODO: implement as contains in the future refactor
    @classmethod
    def has_member(cls, community, user):
        # TODO: maybe int(user_id)
        return CommunityMember.query.filter_by(
            comm_id=community.id, user_id=user.id).one_or_none()

    @classmethod
    def has_admin(cls, community, user):
        # TODO: maybe int(user_id)
        return CommunityMember.query.filter_by(
            comm_id=community.id, user_id=user.id, role='A').one_or_none()


class MembershipRequestAPI(object):

    @classmethod
    def get_invitation(cls, membership_request_id):
        request = MembershipRequest.query.get(membership_request_id)
        community = CommunityMetadata.query.get(request.comm_id)
        return (request, community)

    @classmethod
    def accept_invitation(cls, request, role=None, user_id=None):
        if request.is_invite and not request.user_id:
            request.user_id = user_id
        else:  # add check for permissions on who is handling requesests
            request.role = role
        community_member = CommunityMember.create_from_object(request)
        return community_member

    @classmethod
    def decline_or_cancel_invitation(cls, membership_request_id):
        MembershipRequest.delete(membership_request_id)
        pass

    @classmethod
    def modify_membership_request(cls, comm_id, email, role):
        existing_membership_req = MembershipRequest.query.filter_by(
                comm_id=comm_id,
                email=email
                ).one_or_none()
        if existing_membership_req:
            existing_membership_req.role = role
            return existing_membership_req
        else:
            return None

    @classmethod
    def get_community_outgoing_requests(
            cls, comm_id, page_size=20, page_number=0):
        incoming_count = MembershipRequest.query.filter_by(
            is_invite=True, comm_id=comm_id).count()
        incoming_requests = MembershipRequest.query.filter_by(
            is_invite=True, comm_id=comm_id)[page_number*page_size:page_size]
        return (incoming_count, incoming_requests)

    @classmethod
    def get_community_incoming_requests(
            cls, comm_id, page_size=20, page_number=0):
        outgoing_count = MembershipRequest.query.filter_by(
            is_invite=False, comm_id=comm_id).count()
        outgoing_requests = MembershipRequest.query.filter_by(
            is_invite=False, comm_id=comm_id)[page_number*page_size:page_size]
        return (outgoing_count, outgoing_requests)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Requests database models."""

from __future__ import absolute_import, print_function

import uuid

from invenio_accounts.models import User
from invenio_db import db
from invenio_records.models import RecordMetadataBase, Timestamp
from sqlalchemy.dialects import mysql
from sqlalchemy_utils.types import UUIDType


# TODO: Move to invenio-requests
class Request(db.Model, RecordMetadataBase):
    """Requests model."""

    __tablename__ = 'requests_request'
    __table_args__ = {'extend_existing': True}
    __versioned__ = {'versioning': False}

    owner_id = db.Column(
        db.Integer,
        db.ForeignKey(User.id),
    )

    expires_at = db.Column(
        db.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
    )

    cancelled_at = db.Column(
        db.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
    )

    sla = db.Column(
        db.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
    )

    routing_key = db.Column(
        db.String(255),
    )

    owner = db.relationship(
        User,
        backref='requests',
    )

    @classmethod
    def delete(cls, request):
        """Delete request."""
        with db.session.begin_nested():
            db.session.delete(request)


class Comment(db.Model, Timestamp):
    """Comment in a request."""

    __tablename__ = 'requests_comment'

    id = db.Column(
        UUIDType,
        default=uuid.uuid4,
        primary_key=True,
    )

    request_id = db.Column(
        UUIDType,
        db.ForeignKey(Request.id, ondelete="CASCADE"),
        nullable=False,
    )

    created_by = db.Column(
        db.Integer,
        db.ForeignKey(User.id),
        nullable=False
    )

    message = db.Column(db.Text)

    request = db.relationship(
        Request,
        backref=db.backref(
            'comments',
            cascade='all, delete-orphan',
            passive_deletes=True,
        ),
    )

    @classmethod
    def create(cls, request_id, created_by, message):
        with db.session.begin_nested():
            obj = cls(
                request_id=request_id,
                created_by=created_by,
                message=message,
            )
            db.session.add(obj)
        return obj



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
from flask import request
from flask_security import current_user
from invenio_records_rest.loaders.marshmallow import MarshmallowErrors
from marshmallow import ValidationError
from marshmallow import __version_info__ as marshmallow_version


def marshmallow_loader(schema_class):
    """Marshmallow loader for JSON requests."""
    def json_loader():
        request_json = request.get_json()
        context = {}
        pid_data = request.view_args.get('pid_value')
        if pid_data:
            pid, record = pid_data.data
            context['pid'] = pid
            context['record'] = record
        if current_user.is_authenticated:
            context['user_id'] = int(current_user.get_id())
        if marshmallow_version[0] < 3:
            result = schema_class(context=context).load(request_json)
            if result.errors:
                raise MarshmallowErrors(result.errors)
        else:
            # From Marshmallow 3 the errors on .load() are being raised rather
            # than returned. To adjust this change to our flow we catch these
            # errors and reraise them instead.
            try:
                result = schema_class(context=context).load(request_json)
            except ValidationError as error:
                raise MarshmallowErrors(error.messages)

        return result.data
    return json_loader



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community database models."""

from __future__ import absolute_import, print_function

from invenio_db import db
from invenio_records.models import RecordMetadataBase


class CommunityMetadata(db.Model, RecordMetadataBase):
    """Represent a community."""

    __tablename__ = 'communities_metadata'
    __table_args__ = {'extend_existing': True}
    __versioned__ = {'versioning': False}

    is_deleted = db.Column(
        db.Boolean(name="ck_communities_community_metadata_is_deleted"),
        nullable=True,
        default=False
    )
    """Was the community soft-deleted."""

    def delete(self):
        """Mark the community for deletion."""
        self.is_deleted = True



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Indexer for communities."""

from invenio_indexer.api import RecordIndexer

from .api import Community


class CommunityIndexer(RecordIndexer):
    """Community indexer."""

    record_cls = Community




# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community records errors."""


class CommunityRecordError(Exception):
    """Community record base error class."""

    def __init__(self, community_pid_id, record_pid_id):
        """Initialize Exception."""
        self.community_pid_id = community_pid_id
        self.record_pid_id = record_pid_id


class CommunityRecordAlreadyExists(CommunityRecordError):
    """Record inclusion already exists error."""


class CommunityRecordDoesNotExist(CommunityRecordError):
    """Record inclusion does not exist error."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""JSON Schemas."""

from __future__ import absolute_import, print_function

import uuid

from invenio_pidstore.models import PersistentIdentifier
from invenio_records_rest.schemas import Nested, StrictKeysMixin
from invenio_records_rest.schemas.fields import DateString, GenFunction, \
    SanitizedHTML, SanitizedUnicode
from marshmallow import ValidationError, fields, missing, validate

from invenio_communities.api import Community


def pid_from_context_or_rec(data_value, context, **kwargs):
    """Get PID from marshmallow context."""
    pid = (context or {}).get('pid')
    pid_value = getattr(pid, 'pid_value', None) or data_value
    if not pid_value:
        raise ValidationError('Missing data for required field.')
    else:
        if not pid:  # check that the ID is not already taken
            if PersistentIdentifier.query.filter_by(
                    pid_type='comid', pid_value=pid_value).one_or_none():
                raise ValidationError(
                    'ID "{}" is already assigned to a community.'.format(pid_value))
        return pid_value


def load_creator(_, context):
    """Load the record creator."""
    old_data = context.get('record')
    if old_data:
        return old_data.get('created_by', missing)
    # TODO a validation error must be raised in each case
    return context.get('user_id', missing)


def serialize_creator(record, context):
    """Load the record creator."""
    return record.get('created_by', missing)


class CommunitySchemaMetadataV1(StrictKeysMixin):
    """Community metadata schema."""

    schema_ = fields.Str(attribute="$schema", dump_to="$schema")

    id = GenFunction(
        deserialize=pid_from_context_or_rec,
        serialize=pid_from_context_or_rec  # to be added only when loading
    )
    title = SanitizedUnicode(required=True)
    description = SanitizedHTML()
    curation_policy = SanitizedHTML()
    page = SanitizedHTML()
    type = fields.Str(required=True, validate=validate.OneOf([
        'organization',
        'event',
        'topic',
        'project',
    ]))
    alternate_identifiers = fields.List(fields.Raw())
    website = fields.Url()
    funding = fields.List(fields.String())
    domain = fields.List(fields.String())
    verified = fields.Boolean()
    visibility = fields.Str(validate=validate.OneOf([
        'public',
        'private',
        'hidden',
    ]))
    member_policy = fields.Str(validate=validate.OneOf([
        'open',
        'closed',
    ]))
    record_policy = fields.Str(validate=validate.OneOf([
        'open',
        'closed',
        'restricted',
    ]))
    archived = fields.Boolean()
    created_by = GenFunction(
        deserialize=load_creator,
        serialize=serialize_creator
    )


class CommunitySchemaV1(StrictKeysMixin):
    """Schema for the community metadata."""

    created = fields.Str(dump_only=True)
    revision = fields.Integer(dump_only=True)
    updated = fields.Str(dump_only=True)
    links = fields.Raw(dump_only=True)
    metadata = fields.Nested(CommunitySchemaMetadataV1)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Blueprint definitions."""

from __future__ import absolute_import, print_function

from flask import Blueprint, abort, jsonify, render_template, request
from flask.views import MethodView
from flask_login import current_user, login_required
from invenio_db import db
from webargs import fields, validate

from invenio_communities.records.api import CommunityRecordsCollection

from ..utils import comid_url_converter
from ..views import pass_community, use_kwargs
from .api import CommunityMembersAPI, MembershipRequestAPI
from .models import CommunityMember, MembershipRequest


def create_blueprint_from_app(app):
    """Blueprint creating function."""
    community_members_rest_blueprint = Blueprint(
        'invenio_communities_members',
        __name__,
        template_folder="../templates",
    )
    comm_view = CommunityMembersResource.as_view(
        'community_members_api'
    )
    community_members_rest_blueprint.add_url_rule(
        '/communities/<{}:pid_value>/members'.format(comid_url_converter),
        view_func=comm_view,
    )

    request_management_view = CommunityRequestsResource.as_view(
        'community_requests_management_api'

    )
    community_members_rest_blueprint.add_url_rule(
        '/communities/<{}:pid_value>'
        '/members/requests'.format(comid_url_converter),
        view_func=request_management_view,
    )

    request_view = MembershipRequestResource.as_view(
        'community_requests_api'
    )
    community_members_rest_blueprint.add_url_rule(
        '/communities/members/requests/<membership_request_id>',
        view_func=request_view,
    )
    return community_members_rest_blueprint


class CommunityMembersResource(MethodView):
    """Resource for creating, listing and removing community memberships."""

    post_args = {
        'role': fields.Raw(
            location='json',
            required=False,
            validate=[validate.OneOf(['M', 'A', 'C'])]
        ),
        'email': fields.Email(
            location='json',
            required=False
        ),
        'request_type': fields.Raw(
            location='json',
            required=True,
            validate=[validate.OneOf(['invitation', 'request'])]
        )

    }

    put_args = {
        'role': fields.Raw(
            location='json',
            required=False,
            validate=[validate.OneOf(['M', 'A', 'C'])]
        ),
        'user_id': fields.Raw(
            location='json',
            required=False
        ),
        'email': fields.Email(
            location='json',
            required=False
        ),
    }

    delete_args = {
        'user_id': fields.Raw(
            location='query',
            required=False
        ),
    }

    @use_kwargs(post_args)
    @pass_community
    def post(self, comid=None, community=None, email=None, role=None,
             request_type=None, **kwargs):
        """Join a community or invite a user to it."""
        if request_type == 'invitation':
            admin_ids = \
                [admin.user.id for admin in CommunityMember.get_admins(
                    community.id)]
            if int(current_user.get_id()) not in admin_ids:
                abort(404)
            existing_membership_req = MembershipRequest.query.filter_by(
                    comm_id=community.id,
                    email=email
                    ).one_or_none()
            if existing_membership_req:
                abort(400, 'This is an already existing relationship.')
            CommunityMembersAPI.invite_member(community, email, role)
        elif request_type == 'request':
            user_id = int(current_user.get_id())
            email = current_user.email
            CommunityMembersAPI.join_community(user_id, email, community)
        db.session.commit()
        return 'Succesfully Invited', 200

    @pass_community
    def get(self, comid=None, community=None):
        """List the community members."""
        admin_ids = \
            [admin.user.id for admin in CommunityMember.get_admins(
                community.id)]
        if int(current_user.get_id()) not in admin_ids:
            abort(404)
        ui_members = []
        for member in CommunityMembersAPI.get_members(community.id).all():
            add_member = {}
            add_member['user_id'] = str(member.user_id)
            add_member['email'] = member.user.email
            add_member['role'] = str(member.role.title)
            ui_members.append(add_member)
        return jsonify(ui_members), 200

    @use_kwargs(delete_args)
    @pass_community
    def delete(self, comid=None, community=None, user_id=None):
        """Remove a member from a community."""
        admin_ids = \
            [admin.user.id for admin in CommunityMember.get_admins(
                community.id)]
        if int(current_user.get_id()) not in admin_ids:
            abort(404)
        CommunityMembersAPI.delete_member(community.id, user_id)
        db.session.commit()
        return 'Succesfully removed', 204


class CommunityRequestsResource(MethodView):
    """Resource to list community membership requests."""

    post_args = {
        'response': fields.Raw(
            location='json',
            required=True,
            validate=[validate.OneOf(['accept', 'decline'])]
        ),
        'role': fields.Raw(
            location='json',
            required=False,
            validate=[validate.OneOf(['M', 'A', 'C'])]
        )
    }

    @pass_community
    def get(self, comid=None, community=None, outgoing_only=False,
            incoming_only=False, page_size=20, page=0):
        """List all the community membership requests."""
        admin_ids = \
            [admin.user.id for admin in CommunityMember.get_admins(
                community.id)]
        if int(current_user.get_id()) not in admin_ids:
            abort(404)
        response_object = {}
        if not outgoing_only:
            count, requests = \
                MembershipRequestAPI.get_community_incoming_requests(
                    community.id, page_size=page_size)
            response_object['inc_count'] = count
            response_object['inc_requests'] = []
            for req in requests:
                response_object['inc_requests'].append({
                    'email': req.email,
                    'req_id': req.id
                })
        if not incoming_only:
            count, requests = \
                MembershipRequestAPI.get_community_outgoing_requests(
                    community.id, page_size=page_size)
            response_object['out_count'] = count
            response_object['out_requests'] = []
            for req in requests:
                response_object['out_requests'].append({
                    'email': req.email,
                    'req_id': req.id,
                    'role': str(req.role.title)
                })
        return jsonify(response_object), 200


class MembershipRequestResource(MethodView):
    """Resource to view and handle membership requests."""

    post_args = {
        'response': fields.Raw(
            location='json',
            required=True,
            validate=[validate.OneOf(['accept', 'decline'])]
        ),
        'role': fields.Raw(
            location='json',
            required=False,
            validate=[validate.OneOf(['M', 'A', 'C'])]
        )
    }

    put_args = {
        'role': fields.Raw(
            location='json',
            required=True,
            validate=[validate.OneOf(['M', 'A', 'C'])]
        )
    }

    def get(self, membership_request_id):
        """Get the information for a membership request."""
        member_request, community = MembershipRequestAPI.get_invitation(
            membership_request_id)
        response_object = {}
        response_object['community_name'] = community.json['title']
        response_object['community_id'] = community.json['id']
        response_object['role'] = str(member_request.role.title)
        return jsonify(response_object), 200

    @use_kwargs(post_args)
    def post(self, membership_request_id, response=None, role=None):
        """Accept or reject a membership request."""
        if not current_user.is_authenticated:
            abort(404)
        user_id = int(current_user.get_id())
        member_request = MembershipRequest.query.get(membership_request_id)
        if not member_request.is_invite:
            # and current_user.email == request.email):
            community_admins = [admin.user.id for admin in
                                CommunityMember.get_admins(
                                    member_request.comm_id)]
            if not (user_id in community_admins and not
                    member_request.is_invite):
                abort(404)
        if response == 'accept':
            MembershipRequestAPI.accept_invitation(
                request, role, user_id)
            db.session.commit()
            return 'Succesfully accepted.', 200
        else:
            MembershipRequestAPI.decline_or_cancel_invitation(
                member_request.id)
            db.session.commit()
            return 'Succesfully rejected.', 200

    @use_kwargs(put_args)
    def put(self, membership_request_id, role=None):
        """Modify a membership request."""
        if not current_user.is_authenticated:
            abort(404)
        member_request = MembershipRequest.query.get(membership_request_id)
        community_admins = [admin.user.id for admin in
                            CommunityMember.get_admins(member_request.comm_id)]
        if not (int(current_user.get_id()) in community_admins and
                member_request.is_invite):
            abort(404)
        member_request.role = role
        db.session.commit()
        return 'Succesfully modified invitaion.', 200

    def delete(self, membership_request_id):
        """Cancel(remove) a membership request."""
        if not current_user.is_authenticated:
            abort(404)
        user_id = int(current_user.get_id())
        member_request = MembershipRequest.query.get(membership_request_id)
        if not (current_user.email == member_request.email and not
                member_request.is_invite):
            community_admins = [admin.user.id for admin in
                                CommunityMember.get_admins(
                                    member_request.comm_id)]
            if not (user_id in community_admins and member_request.is_invite):
                abort(404)
        MembershipRequestAPI.decline_or_cancel_invitation(member_request.id)

        db.session.commit()
        return 'Succesfully cancelled invitation.', 204


ui_blueprint = Blueprint(
    'invenio_communities_members',
    __name__,
    template_folder='../templates',
)


@login_required
@ui_blueprint.route(
    '/communities/<{}:pid_value>/members'.format(comid_url_converter))
@pass_community
def members(comid=None, community=None):
    """Members of a community."""
    pending_records = \
        len(CommunityRecordsCollection(community).filter({'status': 'P'}))

    return render_template(
        'invenio_communities/members.html',
        community=community,
        comid=comid,
        pending_records=pending_records
        )


@login_required
@ui_blueprint.route('/communities/members/requests/<membership_request_id>')
def requests(membership_request_id):
    """Requests of communities."""
    return render_template('invenio_communities/request.html')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Record database models."""

from __future__ import absolute_import, print_function

from enum import Enum

from flask_babelex import gettext
from invenio_db import db
from invenio_pidstore.models import PersistentIdentifier
from invenio_records.models import RecordMetadataBase
from speaklater import make_lazy_gettext
from sqlalchemy.exc import IntegrityError
from sqlalchemy_utils.types import ChoiceType, UUIDType

from invenio_communities.models import CommunityMetadata
from invenio_communities.requests.models import Request

from .errors import CommunityRecordAlreadyExists

# TODO make sure well what this does and that we need this dependency
_ = make_lazy_gettext(lambda: gettext)


COMMUNITY_RECORD_STATUS = {
    'PENDING': _('Pending'),
    'ACCEPTED': _('Accepted'),
    'REJECTED': _('Rejected')
}


class CommunityRecordStatus(Enum):
    """Community-record relationship status."""

    PENDING = 'P'
    ACCEPTED = 'A'
    REJECTED = 'R'

    @property
    def title(self):
        """Return human readable title."""
        return COMMUNITY_RECORD_STATUS[self.name]


class CommunityRecord(db.Model, RecordMetadataBase):
    """Comunity-record relationship model."""

    __tablename__ = 'communities_community_record'
    __table_args__ = (
        db.Index(
            'uidx_community_pid_record_pid',
            'community_pid_id', 'record_pid_id',
            unique=True,
        ),
        {'extend_existing': True},
    )
    __versioned__ = {'versioning': False}

    community_pid_id = db.Column(
        db.Integer,
        db.ForeignKey(PersistentIdentifier.id),
        nullable=False,
    )

    record_pid_id = db.Column(
        db.Integer,
        db.ForeignKey(PersistentIdentifier.id),
        nullable=False,
    )

    request_id = db.Column(
        UUIDType,
        db.ForeignKey(Request.id),
        # TODO: should we also allow CommunityRecords without a request?
        nullable=False,
    )

    status = db.Column(
        ChoiceType(CommunityRecordStatus, impl=db.CHAR(1)),
        nullable=False,
        default=CommunityRecordStatus.PENDING,
    )

    community_pid = db.relationship(
        PersistentIdentifier,
        foreign_keys=[community_pid_id],
    )
    record_pid = db.relationship(
        PersistentIdentifier,
        foreign_keys=[record_pid_id],
    )

    request = db.relationship(Request)

    @property
    def commmunity(self):
        """Return community model."""
        # TODO: make a JOIN instead?
        return CommunityMetadata.query.get(self.community_pid.object_uuid)

    @classmethod
    def create(cls, community_pid_id, record_pid_id, request_id,
               status=None, json=None):
        try:
            with db.session.begin_nested():
                # TODO: check if status None works with default
                obj = cls(
                    community_pid_id=community_pid_id,
                    record_pid_id=record_pid_id,
                    request_id=request_id,
                    status=status,
                    json=json,
                )
                db.session.add(obj)
        # TODO: Check if actually this constraint check happens on the DB side
        #       when db.session.add() is called.
        except IntegrityError:
            raise CommunityRecordAlreadyExists(
                community_pid_id=community_pid_id,
                record_pid_id=record_pid_id,
            )
        return obj

    @classmethod
    def delete(cls, community_record):
        """Delete community record relationship."""
        with db.session.begin_nested():
            db.session.delete(community_record)

    # TODO: investigate using RECORDS_PIDS_OBJECT_TYPES or
    #       current_pidstore.resolve
    # @property
    # def record(self)
    #     pass
    #
    # RECORDS_PIDS_OBJECT_TYPES = {
    #     'rec': RecordMetadata,
    #     'com': CommunityMetadata,
    # }



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Blueprint definitions for records integration."""

import uuid

from flask import Blueprint, abort, jsonify, render_template
from flask.views import MethodView
from flask_login import current_user, login_required
from invenio_db import db
from invenio_indexer.api import RecordIndexer
from sqlalchemy.exc import SQLAlchemyError
from webargs import fields

from invenio_communities.records.api import CommunityInclusionRequest, \
    CommunityRecord, CommunityRecordsCollection, Record, \
    RecordCommunitiesCollection

from ..utils import comid_url_converter
from ..views import pass_community, use_kwargs
from .errors import CommunityRecordAlreadyExists

#
# UI views
#
ui_blueprint = Blueprint(
    'invenio_communities_records',
    __name__,
    template_folder='../templates',
)


@ui_blueprint.route(
    '/communities/<{}:pid_value>/curate'.format(comid_url_converter))
@login_required
@pass_community
def curation(comid=None, community=None):
    """Create a new community."""
    pending_records = \
        len(CommunityRecordsCollection(community).filter({'status': 'P'}))

    return render_template(
        'invenio_communities/curation.html',
        community=community,
        comid=comid,
        pending_records=pending_records
    )


#
# REST API views
#
api_blueprint = Blueprint(
    'invenio_communities_records',
    __name__,
    template_folder='../templates',
)


class ListResource(MethodView):
    """Resource for listing and creating inclusion requests."""

    post_args = {
        'record_pid': fields.String(
            location='json',
            required=True,
        ),
        'message': fields.String(
            location='json',
            required=False,
        ),
        'auto_accept': fields.Boolean(
            location='json',
            required=False,
        )
    }

    @pass_community
    @login_required
    def get(self, comid=None, community=None):
        """List the community records requests."""
        # TODO: check if user in community or record owner
        json_response = community.records.as_dict()
        return jsonify(json_response), 200

    @use_kwargs(post_args)
    @pass_community
    @login_required
    def post(self, comid=None, community=None, record_pid=None,
             message=None, auto_accept=False, **kwargs):
        """Join a community or invite a user to it."""
        #
        # View - context parsing/building
        #
        user = current_user

        #
        # Controller
        #
        # TODO: implement "__contains__" in members collection API
        # admin_ids = \
        #     [admin.user.id for admin in CommunityMember.get_admins(
        #         community.id)]
        # if int(current_user.get_id()) not in admin_ids:
        #     abort(404)

        # TODO: Consider also
        # "CommunityInclusionRequest.community_record_cls.record_cls.resolve"
        record_pid, record = CommunityRecord.record_cls.resolve(record_pid)
        request = CommunityInclusionRequest.create(user, id_=uuid.uuid4())
        if message:
            request.add_comment(user, message)
        try:
            com_rec = CommunityRecord.create(
                record, record_pid, community, request)
        except CommunityRecordAlreadyExists:
            abort(400, 'Community-record relationship already exists.')
        com_rec.status = com_rec.Status.PENDING

        # TODO: Check in community members instead of just creator
        # if CommunityMembersAPI.has_member(community, user):
        if community['created_by'] == user.id:
            if auto_accept:
                # TODO: define what the routing_key value should be
                # request.routing_key = ''
                # TODO: implement state as a "Request.state" property?
                request['state'] = request.State.CLOSED
                com_rec.status = com_rec.Status.ACCEPTED
            else:
                # TODO: define what the routing_key value should be
                # TODO: don't use f-strings
                # request.routing_key = f'record:{record_pid.pid_value}:owners'
                pass
        else:
            # TODO: define what the routing_key value should be
            # request.routing_key = \
            #     f'community:{community.pid.pid_value}:curators'
            pass
        db.session.commit()

        # Notify request owners and receivers
        # TODO: implement mail sending (e.g. via "routing_key")
        # send_request_emails(request)

        # Index record with new inclusion request info
        # TODO: Implement indexer receiver to include community info in record
        RecordIndexer().index_by_id(record.id)

        json_response = request.as_dict()
        json_response['links'] = request.dump_links(request, comid.pid_value)

        #
        # View - serialize response
        #
        # TODO: Use marshmallow to serialize
        return jsonify(json_response), 201


class ItemResource(MethodView):
    """Community inclusion request item endpoint."""

    @pass_community
    @login_required
    def get(self, comid=None, community=None, request_id=None):
        """Get the inclusion request."""
        # TODO: check if user in community or record owner
        request = CommunityInclusionRequest.get_record(request_id)
        json_response = request.as_dict()
        return jsonify(json_response), 200

    @pass_community
    @login_required
    def delete(self, comid=None, community=None, request_id=None):
        """Delete the inclusion request."""
        # TODO: check if user in community or record owner
        # TODO: make a "pass_request" decorator
        try:
            request = CommunityInclusionRequest.get_record(request_id)
        except SQLAlchemyError:
            abort(404, 'Record inclusion request not found.')
        community_record = request.community_record
        community_record.delete()
        request.delete()
        db.session.commit()
        return 204


# TODO: move things around
class ItemActionsResource(MethodView):
    """Perform actions on inclusion requests."""

    post_args = {
        'message': fields.String(
            location='json',
            required=False,
        )
    }

    @use_kwargs(post_args)
    @pass_community
    @login_required
    def post(self, comid=None, community=None, request_id=None, action=None,
             message=None, **kwargs):
        """Handle a community record request."""
        #
        # View - context parsing/building
        #
        user = current_user

        #
        # Controller
        #
        # admin_ids = \
        #     [admin.user.id for admin in CommunityMember.get_admins(
        #         community.id)]
        # if int(current_user.get_id()) not in admin_ids:
        #     abort(404)
        # TODO: implement as a "pass_community_request" decorator?
        request = CommunityInclusionRequest.get_record(request_id)
        community_record = request.community_record
        if action == 'comment':
            if not message:
                # TODO: add message missing
                abort(400, 'Missing comment message.')
        elif action == 'accept':
            community_record.status = community_record.Status.ACCEPTED
            # TODO: figure out "routing_key"
            # request.routing_key = ''
            # TODO: use ".state" property setter/getter
            request['state'] = request.State.CLOSED
        elif action == 'reject':
            community_record.status = community_record.Status.REJECTED
            # TODO: figure out "routing_key"
            # request.routing_key = ''
            request['state'] = request.State.CLOSED
        else:
            # TODO: no appropriate action error
            abort(404)
        if message:
            request.add_comment(user, message)
        db.session.commit()

        # Notify request owners and receivers
        # TODO: implement mail sending
        # send_request_emails(request)

        # Index record with new inclusion request info
        # TODO: Implement indexer receiver to include community info in record
        # TODO: check if it's always needed to index the record
        RecordIndexer().index_by_id(community_record.record.id)

        #
        # View - serialize response
        #
        # TODO: User marshmallow to serialize
        return jsonify(community_record.as_dict()), 201


api_blueprint.add_url_rule(
    '/communities/<{pid}:pid_value>'
    '/requests/inclusion'.format(pid=comid_url_converter),
    view_func=ListResource.as_view('requests_list'),
)

api_blueprint.add_url_rule(
    '/communities/<{pid}:pid_value>'
    '/requests/inclusion/<uuid:request_id>'.format(pid=comid_url_converter),
    view_func=ItemResource.as_view('requests_item'),
)

api_blueprint.add_url_rule(
    '/communities/<{pid}:pid_value>'
    '/requests/inclusion/<uuid:request_id>'
    '/<any({actions}):action>'.format(
        pid=comid_url_converter,
        actions=','.join(['accept', 'reject', 'comment'])),
    view_func=ItemActionsResource.as_view('requests_item_actions'),
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Blueprint definitions for communities core."""

from __future__ import absolute_import, print_function

from functools import partial, wraps

from flask import Blueprint, render_template, request
from flask_login import login_required
from flask_menu import current_menu
from invenio_records_rest.errors import PIDResolveRESTError
from invenio_rest.errors import FieldError, RESTValidationError
from sqlalchemy.exc import SQLAlchemyError
from webargs import ValidationError
from webargs.flaskparser import FlaskParser as FlaskParserBase

from invenio_communities.records.api import CommunityRecordsCollection

from .utils import comid_url_converter


class FlaskParser(FlaskParserBase):
    """Parser to add FieldError to validation errors."""

    def handle_error(self, error, *args, **kwargs):
        """Handle errors during parsing."""
        if isinstance(error, ValidationError):
            _errors = []
            for field, messages in error.messages.items():
                _errors.extend([FieldError(field, msg) for msg in messages])
            raise RESTValidationError(errors=_errors)
        super(FlaskParser, self).handle_error(error, *args, **kwargs)


webargs_parser = FlaskParser()
use_args = webargs_parser.use_args
use_kwargs = webargs_parser.use_kwargs


def pass_community(func=None, view_arg_name='pid_value'):
    """Decorator to retrieve persistent identifier and community."""
    if func is None:
        return partial(pass_community, view_arg_name=view_arg_name)

    @wraps(func)
    def inner(*args, **kwargs):
        try:
            del kwargs[view_arg_name]
            value = request.view_args[view_arg_name]
            comid, community = value.data
        except SQLAlchemyError:
            raise PIDResolveRESTError(value)
        return func(*args, comid=comid, community=community, **kwargs)
    return inner


ui_blueprint = Blueprint(
    'invenio_communities',
    __name__,
    template_folder='templates',
    static_folder='static',
)


@ui_blueprint.before_app_first_request
def init_menu():
    """Initialize menu before first request."""
    item = current_menu.submenu('main.communities')
    item.register(
        'invenio_communities.index',
        'Communities',
        order=3,
    )


@ui_blueprint.route('/communities')
def index():
    """Search communities."""
    return render_template('invenio_communities/index.html')


@ui_blueprint.route('/communities/new')
@login_required
def new():
    """Create a new community."""
    return render_template('invenio_communities/new.html')


@ui_blueprint.route(
    '/communities/<{pid}:pid_value>'.format(pid=comid_url_converter))
@pass_community
def community_page(comid=None, community=None):
    """Members of a community."""
    pending_records = \
        len(CommunityRecordsCollection(community).filter({'status': 'P'}))
    return render_template(
        'invenio_communities/community_page.html',
        community=community,
        comid=comid,
        pending_records=pending_records
    )


@ui_blueprint.route(
    '/communities/<{pid}:pid_value>/settings'.format(pid=comid_url_converter))
@pass_community
@login_required
def settings(comid=None, community=None):
    """Modify a community."""
    pending_records = \
        len(CommunityRecordsCollection(community).filter({'status': 'P'}))
    return render_template(
        'invenio_communities/settings.html',
        community=community,
        comid=comid,
        pending_records=pending_records)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Records API."""

from __future__ import absolute_import, print_function

from flask import current_app
from invenio_db import db
from invenio_jsonschemas import current_jsonschemas
from invenio_pidstore.models import PersistentIdentifier
from invenio_pidstore.resolver import Resolver
from invenio_records.api import Record
from werkzeug.local import LocalProxy

from invenio_communities.members.models import CommunityMetadata

from .members.api import CommunityMembersAPI


# TODO: Move somewhere appropriate (`invenio-records-pidstore`)
class PIDRecordMixin:
    """Persistent identifier mixin for records."""

    object_type = None
    pid_type = None

    @property
    def pid(self):
        """Return primary persistent identifier of the record."""
        return PersistentIdentifier.query.filter_by(
            object_uuid=self.id,
            object_type=self.object_type,
            pid_type=self.pid_type
        ).one()

    @classmethod
    def resolve(cls, pid_value):
        """Resolve a PID value and return the PID and record."""
        return Resolver(
            pid_type=cls.pid_type,
            object_type=cls.object_type,
            getter=cls.get_record
        ).resolve(pid_value)

    # TODO: See if needed or if it should be customizable
    # @property
    # def pids(self):
    #     """Return all persistent identifiers of the record."""
    #     return PersistentIdentifier.query.filter_by(
    #         object_uuid=self.id,
    #         object_type=self.object_type,
    #     ).all()


class CommunityBase(Record, PIDRecordMixin):
    """Define API for community creation and manipulation."""

    object_type = 'com'
    pid_type = 'comid'

    # TODO: Communities model doesn't have versioninig, some methods from
    # "invenio_records.api.RecordBase" have to be overridden/removed
    model_cls = CommunityMetadata

    schema = LocalProxy(lambda: current_jsonschemas.path_to_url(
        current_app.config.get(
            'COMMUNITY_SCHEMA', 'communities/communities-v1.0.0.json')))

    @classmethod
    def create(cls, data, id_=None, **kwargs):
        """Create a new community instance and store it in the database."""
        with db.session.begin_nested():
            data['$schema'] = str(cls.schema)
            community = cls(data)
            community.validate(**kwargs)
            community.model = cls.model_cls(id=id_, json=community)
            db.session.add(community.model)
            # TODO: Move this logic to the controller
            CommunityMembersAPI.set_default_admin(community.model)
        return community

    def clear(self):
        """Clear but preserve the schema field."""
        schema = self['$schema']
        super(CommunityBase, self).clear()
        self['$schema'] = schema

    def delete(self, force=False):
        """Delete a community."""
        with db.session.begin_nested():
            if force:
                db.session.delete(self.model)
            else:
                self.model.delete()
        return self


# TODO: Investigate if there's a cleaner/better way
Community = LocalProxy(
    lambda: current_app.extensions['invenio-communities'].community_cls)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Records API."""

from __future__ import absolute_import, print_function

from collections import defaultdict
from enum import Enum

from flask import url_for
from invenio_db import db
from invenio_records.api import Record as RecordBaseAPI
from werkzeug.local import LocalProxy

from invenio_communities.api import Community, PIDRecordMixin
from invenio_communities.records.models import \
    CommunityRecord as CommunityRecordModel
from invenio_communities.records.models import CommunityRecordStatus
from invenio_communities.requests.api import RequestBase


# TODO: See if this can be moved to config
class Record(RecordBaseAPI, PIDRecordMixin):
    """PID-aware record."""

    object_type = 'rec'
    pid_type = 'recid'


class CommunityInclusionRequest(RequestBase):

    TYPE = 'community-inclusion-request'

    # TODO: see how to implement or if needed at all...
    community_record_cls = LocalProxy(lambda: CommunityRecord)

    # TODO: Override
    schema = {
        "type": {
            "type": "string",
            # "enum": ["community-inclusion"],
        },
        "state": {
            "type": "string",
            # "enum": ["pending", "closed"],
        },
        "assignees": {"type": "int[]"},
        "created_by": {"type": "int"},
    }

    # TODO: implement "state" property with getter/setter?
    class State(Enum):
        OPEN = 'open'
        CLOSED = 'closed'

    @property
    def community_record(self):
        """Get request's community record relatinship."""
        if not getattr(self, '_community_record', None):
            self._community_record = \
                self.community_record_cls.get_by_request_id(request_id=self.id)
        return self._community_record

    @property
    def community(self):
        """Get request community."""
        return self.community_record.community

    @property
    def record(self):
        """Get request record."""
        return self.community_record.record

    @classmethod
    def create(cls, owner, id_=None, **kwargs):
        """Create a community inclusion request."""
        data = {
            'type': cls.TYPE,
            'state': cls.State.OPEN.value,
            'created_by': owner.id,
            **kwargs,
        }
        model = cls.model_cls(
            owner_id=owner.id,
            json=data,
            id=id_,
        )
        db.session.add(model)
        return cls(data, model=model)

    def as_dict(self):
        return {
            'id': self.id,
            'created': self.created,
            'updated': self.updated,
            'comments': [
                {
                    'id': c.id,
                    'message': c.message,
                    'created_by': c.created_by,
                    'created': c.created,
                    'updated': c.updated,
                } for c in self.comments
            ],
        }

    # TODO: Probably part of view/controller
    @classmethod
    def dump_links(cls, request, pid_value):
        actions = ['comment', 'accept', 'reject']
        links = {
            "self": url_for(
                'invenio_communities_records.requests_item',
                pid_value=pid_value,
                request_id=request.id
            )
        }
        for action in actions:
            links[action] = url_for(
                'invenio_communities_records.requests_item_actions',
                pid_value=pid_value,
                request_id=request.id,
                action=action
            )
        return links


class RecordCommunitiesMixin(PIDRecordMixin):

    record_communities_iter_cls = LocalProxy(
        lambda: RecordCommunitiesCollection)

    object_type = 'rec'
    pid_type = 'recid'

    @property
    def communities(self):
        return self.record_communities_iter_cls(self)

    # TODO: Take into account in the controllers
    # def block_community(cls, community):
    #     # TODO: should this be implemented on the level of Request API
    #     # TODO create a CommunityMember relationship restricting this
    #     pass


class CommunityRecord(RecordBaseAPI):
    """Community-record API class."""

    model_cls = CommunityRecordModel
    community_cls = Community
    record_cls = Record

    schema = {
        # TODO: Define schema
    }

    Status = CommunityRecordStatus

    @property
    def request(self):
        """Community record request."""
        # TODO: Return a RequestRecord object, not the SQLAlchemy model
        if self.model:
            return CommunityInclusionRequest(
                self.model.request.json, self.model.request)
        else:
            return None

    @property
    def status(self):
        """Get community record relationship status."""
        return self.model.status if self.model else None

    @status.setter
    def status(self, new_status):
        """Set community record relationship status."""
        self.model.status = new_status

    @property
    def community(self):
        """Get request's community record relatinship."""
        if not getattr(self, '_community', None):
            self._community = self.community_cls.get_record(
                self.model.community_pid.object_uuid)
        return self._community

    @property
    def record(self):
        """Get request's community record relatinship."""
        if not getattr(self, '_record', None):
            self._record = self.record_cls.get_record(
                self.model.record_pid.object_uuid)
        return self._record

    @classmethod
    def create(cls, record, record_pid, community, request, status=None,
               can_curate=False, data=None):
        data = data or {}
        model = CommunityRecordModel.create(
            community_pid_id=community.pid.id,
            record_pid_id=record_pid.id,
            request_id=request.id,
            status=status,
            json=data,
        )
        obj = cls(data, model=model)
        obj._community = community
        obj._record = record
        return obj

    def delete(self):
        """Delete community record relationship."""
        with db.session.begin_nested():
            db.session.delete(self.model)
        return self

    @classmethod
    def get_by_pids(cls, community_pid, record_pid):
        """Get by community and record PIDs."""
        model = CommunityRecordModel.query.filter_by(
            community_pid_id=community_pid.id,
            record_pid_id=record_pid.id,
        ).one_or_none()
        if not model:
            return None
        return cls(model.json, model=model)

    @classmethod
    def get_by_request_id(cls, request_id):
        """Get by request ID."""
        model = CommunityRecordModel.query.filter_by(
            request_id=request_id
        ).one_or_none()
        if not model:
            return None
        return cls(model.json, model=model)

    def as_dict(self, include_request=True):
        res = {
            'status': str(self.status.title),
            'record_pid': self.record.pid.pid_value,
        }
        if include_request:
            res['request'] = self.request.as_dict()
        else:
            res['request_id'] = self.request.id
        return res

    # TODO: sanity check this
    @property
    def record_pid_value(self):
        return self.model.record_pid.pid_value

    # TODO: sanity check this
    @property
    def community_pid_value(self):
        return self.model.community_pid.pid_value


class CommunityRecordsCollectionBase:

    community_record_cls = CommunityRecord

    def __len__(self):
        """Get number of community records."""
        return self._query.count()

    def __iter__(self):
        self._it = iter(self._query)
        return self

    def filter(self, conditions):
        new_query = self._query.filter_by(**conditions)
        return self.__class__(self.community, _query=new_query)

    def __next__(self):
        """Get next community record item."""
        obj = next(self._it)
        return self.community_record_cls(obj.json, model=obj)

    def __getitem__(self, key):
        raise NotImplementedError()


class CommunityRecordsCollection(CommunityRecordsCollectionBase):

    def __init__(self, community, _query=None):
        self.community = community
        # TODO: Make lazier (e.g. via property)
        self._query = _query or CommunityRecordModel.query.filter_by(
            community_pid_id=self.community.pid.id)

    # TODO: implement if needed
    # def __contains__(self, item):
    #     pass

    def __getitem__(self, record):
        """Get a specific community record by record PID."""
        return self.community_record_cls.get_by_pids(
            self.community.pid, record.pid)

    def add(self, record, request):
        return self.community_record_cls.create(
            self.community, record, request)

    def remove(self, record):
        community_record = self[record]
        return community_record.delete()

    def as_dict(self, include_request=False):
        community_records = defaultdict(list)
        for community_record in self:
            status = community_record.status.name.lower()
            community_records[status].append(community_record.as_dict(
                include_request=include_request))
        return community_records


class RecordCommunitiesCollection(CommunityRecordsCollectionBase):

    def __init__(self, record, _query=None):
        self.record = record
        # TODO: Make lazier (e.g. via property)
        self._query = _query or CommunityRecordModel.query.filter_by(
            record_pid_id=self.record.pid.id)

    def filter(self, conditions):
        new_query = self._query.filter_by(**conditions)
        return self.__class__(self.community, _query=new_query)

    def __getitem__(self, community):
        """Get a specific community record."""
        return self.community_record_cls.get_by_pids(
            community.pid, self.record.pid)

    def add(self, community, request):
        return self.community_record_cls.create(
            community, self.record, request)

    def remove(self, community):
        community_record = self[community]
        return community_record.delete()

    def as_dict(self):
        res = defaultdict(list)
        for community_record in self:
            status = community_record.status.name.lower()
            res[status].append({
                'id': community_record.community.pid.object_uuid,
                'comid': community_record.community.pid.pid_value,
                'title': community_record.community['title'],
                # TODO: Add when implemented
                # 'logo': None,
                'request_id': str(community_record.request.id),
                'created_by': community_record.request['created_by'],
            })
        return res


class CommunityRecordsMixin:

    community_records_iter_cls = CommunityRecordsCollection

    @property
    def records(self):
        return self.community_records_iter_cls(self)

    # TODO: Take into account in the controllers
    # @property
    # def notified_members(self):
    #     pass
    # @property
    # def allowed_record_integrators(self):
    #     pass
    # @property
    # def banned_record_integrators(self):
    #     pass



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community database models."""

from __future__ import absolute_import, print_function

import uuid
from enum import Enum

from flask_babelex import gettext
from invenio_accounts.models import User
from invenio_db import db
from speaklater import make_lazy_gettext
from sqlalchemy.exc import IntegrityError
from sqlalchemy_utils.types import ChoiceType, UUIDType

from invenio_communities.models import CommunityMetadata

# TODO make sure well what this does and that we need this dependency
_ = make_lazy_gettext(lambda: gettext)


class CommunityMemberAlreadyExists(Exception):
    """Community membership already exists error."""

    def __init__(self, user_id, pid_id, role):
        """Initialize Exception."""
        self.user_id = user_id
        self.pid_id = pid_id
        self.role = role


class CommunityMemberDoesNotExist(CommunityMemberAlreadyExists):
    """Community membership does not exist error."""

    pass


COMMUNITY_MEMBER_TITLES = {
    'MEMBER': _('Member'),
    'ADMIN': _('Admin'),
    'CURATOR': _('Curator')
}


class CommunityRoles(Enum):
    """Constants for possible roles of any given Community member."""
    # TODO Make roles configurable.
    MEMBER = 'M'
    """Member of the community."""

    ADMIN = 'A'
    """Admin of the community."""

    CURATOR = 'C'
    """Curator of the community."""

    @property
    def title(self):
        """Return human readable title."""
        return COMMUNITY_MEMBER_TITLES[self.name]


class CommunityMember(db.Model):
    """Represent a community member role."""

    __tablename__ = 'communities_members'

    "Community PID ID"
    comm_id = db.Column(
        UUIDType,
        db.ForeignKey(CommunityMetadata.id),
        primary_key=True,
        nullable=False,
    )

    "USER ID"
    user_id = db.Column(
        db.Integer,
        db.ForeignKey(User.id),
        primary_key=True,
        nullable=False,
    )

    role = db.Column(
        ChoiceType(CommunityRoles, impl=db.CHAR(1)), nullable=False)

    user = db.relationship(User, backref='communities')

    community = db.relationship(CommunityMetadata, backref='members')

    @property
    def email(self):
        """Get user email."""
        return self.user.email

    @classmethod
    def create_from_object(cls, membership_request):
        """Create Community Membership Role."""
        with db.session.begin_nested():
            obj = cls.create(
                comm_id=membership_request.comm_id,
                user_id=membership_request.user_id,
                role=membership_request.role)
            db.session.delete(membership_request)
        return obj

    @classmethod
    def create(cls, comm_id, user_id, role):
        """Create Community Membership Role."""
        try:
            with db.session.begin_nested():
                obj = cls(
                    comm_id=comm_id,
                    user_id=user_id,
                    role=role)
                db.session.add(obj)
        except IntegrityError:
            raise CommunityMemberAlreadyExists(
                comm_id=comm_id,
                user_id=user_id,
                role=role)
        return obj

    @classmethod
    def delete(cls, comm_id, user_id):
        """Delete a community membership."""
        try:
            with db.session.begin_nested():
                membership = cls.query.filter_by(
                    comm_id=comm_id, user_id=user_id).one()
                db.session.delete(membership)
        except IntegrityError:
            raise CommunityMemberDoesNotExist(comm_id=comm_id, user_id=user_id)

    @classmethod
    def get_admins(cls, comm_id):
        with db.session.begin_nested():
            return cls.query.filter_by(comm_id=comm_id, role='A').all()

    @classmethod
    def get_members(cls, comm_id):
        with db.session.begin_nested():
            return cls.query.filter_by(comm_id=comm_id)


class MembershipRequest(db.Model):
    """Represent a community member role."""

    __tablename__ = 'communities_membership_request'

    id = db.Column(
        UUIDType,
        primary_key=True,
        default=uuid.uuid4,
    )

    user_id = db.Column(
        db.Integer,
        db.ForeignKey(User.id),
        nullable=True,
    )

    comm_id = db.Column(
        UUIDType,
        db.ForeignKey(CommunityMetadata.id),
        nullable=False,
    )

    email = db.Column(
        db.String(255),
        nullable=True,
    )

    role = db.Column(
        ChoiceType(CommunityRoles, impl=db.CHAR(1)), nullable=True)

    is_invite = db.Column(db.Boolean(name='is_invite'), nullable=False,
                          default=True)

    @classmethod
    def create(cls, comm_id, is_invite, role=None, user_id=None, email=None):
        """Create Community Membership request."""
        try:
            with db.session.begin_nested():
                obj = cls(
                    comm_id=comm_id, user_id=user_id,
                    role=role, is_invite=is_invite, email=email)
                db.session.add(obj)
        except IntegrityError:
            raise CommunityMemberAlreadyExists(
                comm_id=comm_id, user_id=user_id, role=role)
        return obj

    @classmethod
    def delete(cls, id):
        """Delete a community membership request."""
        try:
            with db.session.begin_nested():
                membership = cls.query.get(id)
                db.session.delete(membership)
        except IntegrityError:
            raise CommunityMemberDoesNotExist(id)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Community PID minters."""

from __future__ import absolute_import, print_function

from invenio_pidstore.models import PersistentIdentifier, PIDStatus


def comid_minter(record_uuid, data):
    """Mint a community PID."""
    assert data.get('id')
    return PersistentIdentifier.create(
        pid_type='comid', pid_value=data['id'],
        object_type='com', object_uuid=record_uuid,
        status=PIDStatus.REGISTERED
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community permissions."""

from flask_security import current_user


def allow_logged_in(*args, **kwargs):
    """Permission that checks if the community owner is making the request."""
    def can(self):
        """Check that the current user is the community's owner."""
        return current_user.is_authenticated
    return type('AllowCommunityOwner', (), {'can': can})()



def allow_community_owner(record, *args, **kwargs):
    """Permission that checks if the community owner is making the request."""
    def can(self):
        """Check that the current user is the community's owner."""
        if current_user.is_authenticated:
            return current_user.id == record['created_by']
        return False
    return type('AllowCommunityOwner', (), {'can': can})()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community serializers."""

from __future__ import absolute_import, print_function

from invenio_records_rest.serializers.json import JSONSerializer
from invenio_records_rest.serializers.response import record_responsify, \
    search_responsify

from ..marshmallow import CommunitySchemaV1

# Serializers
# ===========
#: JSON serializer definition.
json_v1 = JSONSerializer(CommunitySchemaV1, replace_refs=True)

# Records-REST serializers
# ========================
#: JSON record serializer for individual records.
json_v1_response = record_responsify(json_v1, 'application/json')
#: JSON record serializer for search results.
json_v1_search = search_responsify(json_v1, 'application/json')

__all__ = (
    'json_v1',
    'json_v1_response',
    'json_v1_search',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio communities extension."""

from __future__ import absolute_import, print_function

from invenio_indexer.signals import before_record_index
from werkzeug.utils import cached_property

from . import config
from .utils import LazyPIDConverter


class InvenioCommunities(object):
    """Invenio extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    @cached_property
    def community_cls(self):
        """Base community API class."""
        # TODO: Refactor
        from .api import CommunityBase
        from .records.api import CommunityRecordsMixin
        return type(
            'Community',
            (CommunityBase, CommunityRecordsMixin),
            {},
        )

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        app.extensions['invenio-communities'] = self

        # TODO: Remove when merged in invenio-records-rest
        app.url_map.converters['lazy_pid'] = LazyPIDConverter

        # TODO: Make configurable or move into separate extension in ".records"
        from invenio_communities.records.indexer import record_indexer_receiver
        before_record_index.dynamic_connect(
            record_indexer_receiver, sender=app, weak=False,
            index=app.config.get(
                'COMMUNITIES_RECORD_INDEX', 'records-record-v1.0.0'),
        )

        @app.context_processor
        def record_context_processors():
            """Jinja context processors for communities record integration."""
            def record_communities(record):
                from invenio_communities.records.api import RecordCommunitiesCollection, Record
                return RecordCommunitiesCollection(Record(record, model=record.model))
            return dict(record_communities=record_communities)


    def init_config(self, app):
        """Initialize configuration.

        Override configuration variables with the values in this package.
        """
        for k in dir(config):
            if k.startswith('COMMUNITIES_'):
                if k == 'COMMUNITIES_REST_ENDPOINTS':
                    # Make sure of registration process.
                    app.config['RECORDS_REST_ENDPOINTS'].update(getattr(
                        config, k))
                app.config.setdefault(k, getattr(config, k))
                if k == 'COMMUNITIES_REST_FACETS':
                    # TODO Might be overriden depending on which package is
                    # initialised first
                    app.config['RECORDS_REST_FACETS'].update(
                        getattr(config, k))
        app.config.setdefault(
            'SUPPORT_EMAIL', getattr(config, 'SUPPORT_EMAIL'))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Default configuration."""

from __future__ import absolute_import, print_function

from invenio_records_rest.facets import terms_filter
from invenio_records_rest.utils import allow_all

from .permissions import allow_community_owner, allow_logged_in
from .utils import comid_url_converter

#: Records REST API endpoints.
COMMUNITIES_REST_ENDPOINTS = dict(
    comid=dict(
        pid_type='comid',
        pid_minter='comid',
        pid_fetcher='comid',
        list_route='/communities/',
        item_route='/communities/<{}:pid_value>'.format(comid_url_converter),
        search_index='communities',
        record_class='invenio_communities.api:Community',
        record_serializers={
            'application/json': (
                'invenio_communities.serializers.json_v1_response'),
        },
        search_serializers={
            'application/json': (
                'invenio_communities.serializers:json_v1_search'),
        },
        record_loaders={
            'application/json': ('invenio_communities.loaders:json_v1'),
        },
        indexer_class='invenio_communities.indexer:CommunityIndexer',
        default_media_type='application/json',
        read_permission_factory_imp=allow_all,
        create_permission_factory_imp=allow_logged_in,
        list_permission_factory_imp=allow_all,
        update_permission_factory_imp=allow_community_owner,
        delete_permission_factory_imp=allow_community_owner,
    ),
)

COMMUNITIES_REST_FACETS = dict(
    communities=dict(
        aggs=dict(
            type=dict(
                terms=dict(field="type"),
            ),
            domain=dict(
                terms=dict(field="domain"),
            )
        ),
        post_filters=dict(
            type=terms_filter('type'),
            domain=terms_filter('domain'),
        )
    )
)

COMMUNITIES_MEMBERSHIP_REQUESTS_CONFIRMLINK_EXPIRES_IN = 1000000

SUPPORT_EMAIL = 'test@email.org'

COMMUNITIES_RECORD_INDEX = 'records-record-v1.0.0'



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Loaders."""

from __future__ import absolute_import, print_function

from invenio_communities.loaders.marshmallow import marshmallow_loader

from ..marshmallow import CommunitySchemaMetadataV1

#: JSON loader using Marshmallow for data validation.
json_v1 = marshmallow_loader(CommunitySchemaMetadataV1)

__all__ = (
    'json_v1',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
from __future__ import absolute_import, print_function

from invenio_pidstore.fetchers import FetchedPID


def comid_fetcher(record_uuid, data):
    """Fetch a community's identifier."""
    return FetchedPID(provider=None, pid_type='comid', pid_value=data['id'])



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Requests API."""

from invenio_records.api import Record as RecordBaseAPI

from .models import Comment as CommentModel
from .models import Request as RequestModel


class RequestBase(RecordBaseAPI):
    """Request API class."""

    model_cls = RequestModel

    schema = {
        "type": {
            "type": "string",
            # "enum": [...],
        },
        "state": {
            "type": "string",
            # "enum": ["pending", "closed"],
        },
        "assignees": {"type": "int[]"},
        "created_by": {"type": "int"},
    }

    @property
    def routing_key(self):
        """Get request routing key."""
        return self.model.routing_key if self.model else None

    @routing_key.setter
    def routing_key(self, new_routing_key):
        """Set request routing key."""
        self.model.routing_key = new_routing_key

    @property
    def comments(self):
        """Request comments."""
        return self.model.comments if self.model else None

    def add_comment(self, user, message):
        """Request comments."""
        # TODO: do we need a comment API Class?
        return CommentModel.create(self.id, user.id, message)

    def delete(self):
        self.model.delete(self.model)
